from collections import defaultdict

import numpy as np
import torch

from ..preprocessors import AudioFeatureExtractor, Tokenizer
from ..utils import Logger, convert_batch_dict_dtype


__all__ = [
    "TextPaddingDataCollator",
    "TextGenerationDataCollator",
    "ImageCaptioningDataCollator",
    "SpeechRecognitionDataCollator",
    "SequenceLabelingDataCollator",
    "CharLevelOCRDataCollator",
]

logger = Logger(__name__)


def _convert_to_batch_dict(dicts_list: list[dict]):
    """
    Convert a list of dicts to a dict of batched values.

    Args:
        dicts_list: A list of dictionaries containing the same set of keys

    Returns:
        A dictionary of the batches
    """
    batch_dict = defaultdict(list)
    for item in dicts_list:
        for key, value in item.items():
            batch_dict[key].append(value)
    batch_dict = dict(batch_dict)
    return batch_dict


class TextPaddingDataCollator:
    """
    A data collator that pads a batch of tokenized inputs.

    Args:
        tokenizer (Tokenizer): A Hezar tokenizer instance.
        padding (str): Specifies padding strategy, either `longest` or `max_length`.
        padding_side (str): Specifies from which side of each tensor to add paddings, either `left` or `right`
        max_length (int): If `padding` is set to `max_length` this must be specified. Forces all tensors to have
            this value as length.
        return_tensors (str): Specifies the dtype of the returning tensors in the batch. (`numpy`, `list`, `torch`)
    """

    def __init__(
        self,
        tokenizer: Tokenizer,
        padding: str = "longest",
        padding_side: str = "right",
        max_length: int = None,
        return_tensors: str = "torch",
    ):
        self.tokenizer = tokenizer
        self.padding = padding
        self.padding_side = padding_side
        self.max_length = max_length
        self.return_tensors = return_tensors

        self.field_to_pad_id_mapping = {
            "token_ids": self.tokenizer.pad_token_id,
            "token_type_ids": self.tokenizer.config.pad_token_type_id,
            "tokens": "",
            "special_tokens_mask": 1,
            "attention_mask": 0,
        }

    def __call__(self, input_batch):
        """
        Add padding to every item in the batch

        Args:
            input_batch: A batch dictionary

        Returns:
            Dict: The same batch dictionary but padded
        """
        input_batch = [convert_batch_dict_dtype(x, dtype="list") for x in input_batch]
        input_batch = _convert_to_batch_dict(input_batch)
        labels = input_batch.pop("labels")
        input_length = self.max_length or max(len(x) for x in input_batch["token_ids"])

        for field, batch in input_batch.items():
            padded_batch = []
            for x in batch:
                if isinstance(x, torch.Tensor):
                    x = x.cpu().numpy().tolist()
                elif isinstance(x, np.ndarray):
                    x = x.tolist()
                difference = input_length - len(x)
                paddings = [self.field_to_pad_id_mapping[field]] * difference
                padded_x = x + paddings if self.padding_side == "right" else paddings + x
                padded_batch.append(padded_x)
            input_batch[field] = padded_batch

        input_batch["labels"] = labels

        input_batch = convert_batch_dict_dtype(input_batch, dtype=self.return_tensors)

        return input_batch


class TextGenerationDataCollator:
    """
    A data collator for text to text generation

    Args:
        tokenizer (Tokenizer): A Hezar tokenizer instance.
        padding (str): Specifies padding strategy, either `longest` or `max_length`.
        padding_side (str): Specifies from which side of each tensor to add paddings, either `left` or `right`
        max_length (int): If `padding` is set to `max_length` this must be specified. Forces all tensors to have
            this value as length.
        labels_max_length (int): Maximum target length for text generation.

    """

    def __init__(
        self,
        tokenizer: Tokenizer,
        padding: str = "longest",
        padding_side: str = "right",
        max_length: int = None,
        labels_max_length: int = None,
    ):
        self.tokenizer = tokenizer
        self.padding = padding
        self.padding_side = padding_side
        self.max_length = max_length
        self.labels_max_length = labels_max_length

    def __call__(self, input_batch):
        """
        Add padding to every item in the batch

        Args:
            input_batch (List[Dict]): A batch dictionary

        Returns:
            Dict: The same batch dictionary but padded
        """
        input_batch = [convert_batch_dict_dtype(x, dtype="list") for x in input_batch]
        input_batch = _convert_to_batch_dict(input_batch)
        padded_batch = self.tokenizer.pad_encoded_batch(
            input_batch,
            padding=self.padding,
            max_length=self.max_length,
            exclude_keys=["labels"],
            return_tensors="torch",
        )
        padded_batch = self.tokenizer.pad_encoded_batch(
            padded_batch,
            padding=self.padding,
            max_length=self.labels_max_length,
            include_keys=["labels"],
            return_tensors="torch",
        )

        return padded_batch


class ImageCaptioningDataCollator:
    """
    Data collator for image captioning.

    Args:
        tokenizer (Tokenizer): A Hezar tokenizer instance.
        padding (str): Specifies padding strategy, either `longest` or `max_length`.
        padding_side (str): Specifies from which side of each tensor to add paddings, either `left` or `right`
        max_length (int): If `padding` is set to `max_length` this must be specified. Forces all tensors to have
            this value as length.
    """

    def __init__(
        self,
        tokenizer: Tokenizer,
        padding: str = "longest",
        padding_side: str = "right",
        max_length: int = None,
    ):
        self.tokenizer = tokenizer
        self.padding = padding
        self.padding_side = padding_side
        self.max_length = max_length

    def __call__(self, input_batch):
        input_batch = _convert_to_batch_dict(input_batch)
        input_batch = self.tokenizer.pad_encoded_batch(
            input_batch,
            padding=self.padding,
            max_length=self.max_length,
            exclude_keys=["pixel_values"],
            return_tensors="torch",
        )
        if isinstance(input_batch["pixel_values"], list):
            if isinstance(input_batch["pixel_values"][0], list):
                input_batch["pixel_values"] = torch.tensor(input_batch["pixel_values"])
            elif isinstance(input_batch["pixel_values"][0], torch.Tensor):
                input_batch["pixel_values"] = torch.stack(input_batch["pixel_values"])
            elif isinstance(input_batch["pixel_values"][0], np.ndarray):
                input_batch["pixel_values"] = torch.stack([torch.from_numpy(x) for x in input_batch["pixel_values"]])

        return input_batch


class SpeechRecognitionDataCollator:
    def __init__(
        self,
        feature_extractor: AudioFeatureExtractor,
        tokenizer: Tokenizer,
        inputs_padding: str = "longest",
        inputs_max_length: int = None,
        labels_padding: str = "longest",
        labels_max_length: int = None,
    ):
        self.feature_extractor = feature_extractor
        self.tokenizer = tokenizer
        self.inputs_padding = inputs_padding
        self.inputs_max_length = inputs_max_length
        self.labels_padding = labels_padding
        self.labels_max_length = labels_max_length

    def __call__(self, input_batch):
        input_batch = _convert_to_batch_dict(input_batch)
        inputs = self.tokenizer.pad_encoded_batch(
            input_batch,
            padding=self.labels_padding,
            max_length=self.labels_max_length,
            exclude_keys=["input_features"],
            return_tensors="torch",
        )

        inputs = self.feature_extractor.pad(
            inputs,
            padding=self.inputs_padding,
            max_length=self.inputs_max_length,
            return_tensors="torch",
        )

        return inputs


class SequenceLabelingDataCollator:
    """
    A data collator for sequence labeling.

    Args:
        tokenizer (Tokenizer): A Hezar tokenizer instance.
        padding (str): Specifies padding strategy, either `longest` or `max_length`.
        padding_side (str): Specifies from which side of each tensor to add paddings, either `left` or `right`.
        label_pad_token_id (int): Token ID for padding labels.
        max_length (int): If `padding` is set to `max_length` this must be specified. Forces all tensors to have
            this value as length.
    """

    def __init__(
        self,
        tokenizer: Tokenizer,
        padding: str = "longest",
        padding_side: str = "right",
        label_pad_token_id: int = -100,
        max_length: int = None,
    ):
        self.tokenizer = tokenizer
        self.padding = padding
        self.padding_side = padding_side
        self.label_pad_token_id = label_pad_token_id
        self.max_length = max_length

    def __call__(self, input_batch):
        """
        Add padding to every item in the batch

        Args:
            input_batch (List[Dict]): A batch dictionary

        Returns:
            Dict: The same batch dictionary but padded
        """
        input_batch = _convert_to_batch_dict(input_batch)
        labels = input_batch["labels"]
        self.tokenizer.config.padding_side = self.padding_side
        input_batch = self.tokenizer.pad_encoded_batch(
            input_batch,
            padding=self.padding,  # noqa
            max_length=self.max_length,
            return_tensors="torch",
        )

        if labels is None:
            return input_batch

        input_batch.pop("word_ids", None)
        sequence_length = input_batch["token_ids"].shape[1]
        if self.padding_side == "right":
            input_batch["labels"] = [
                list(label) + [self.label_pad_token_id] * (sequence_length - len(label)) for label in labels
            ]
        else:
            input_batch["labels"] = [
                [self.label_pad_token_id] * (sequence_length - len(label)) + list(label) for label in labels
            ]

        input_batch = {
            k: torch.tensor(v) if not isinstance(v, torch.Tensor) else v for k, v in input_batch.items()
        }

        return input_batch


class CharLevelOCRDataCollator:
    """
    A data collator for character-level OCR.

    Args:
        pad_token_id (int): Token ID for padding characters.
    """

    def __init__(self, pad_token_id: int = 0):
        self.pad_token_id = pad_token_id

    def __call__(self, input_batch):
        """
        Add padding to character-level OCR data.

        Args:
            input_batch (Dict): Input batch containing pixel values and labels.

        Returns:
            Dict: Padded input batch.
        """
        input_batch = _convert_to_batch_dict(input_batch)

        if not isinstance(input_batch["pixel_values"][0], torch.Tensor):
            input_batch["pixel_values"] = torch.tensor(input_batch["pixel_values"])
        elif isinstance(input_batch["pixel_values"], list) and isinstance(input_batch["pixel_values"][0], torch.Tensor):
            input_batch["pixel_values"] = torch.stack(input_batch["pixel_values"])

        max_length = max(map(len, input_batch["labels"]))
        all_labels = []
        for labels in input_batch["labels"]:
            labels += [self.pad_token_id] * (max_length - len(labels))
            all_labels.append(labels)
        input_batch["labels"] = torch.tensor(all_labels)
        return input_batch



import csv

import datasets
from datasets.tasks import TextClassification


logger = datasets.logging.get_logger(__name__)

_CITATION = """Citation"""
_DESCRIPTION = """Description"""

_DOWNLOAD_URLS = {
    "train": "PUT PATH TO TRAIN CSV FILE HERE (REQUIRED)",
    "dev": "PUT PATH TO DEV CSV FILE HERE (OPTIONAL)",
    "test": "PUT PATH TO TEST CSV FILE HERE (OPTIONAL BUT RECOMMENDED)",
}


class DatasetNameConfig(datasets.BuilderConfig):
    def __init__(self, **kwargs):
        super(DatasetNameConfig, self).__init__(**kwargs)


class DatasetName(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIGS = [
        DatasetNameConfig(
            name="NAME FOR YOUR DATASET",
            version=datasets.Version("DATASET VERSION"),
            description=_DESCRIPTION,
        ),
    ]

    def _info(self):
        text_column = "text"
        label_column = "label"
        # TODO PROVIDE THE LABELS HERE
        label_names = ["negative", "positive", "neutral"]
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {text_column: datasets.Value("string"), label_column: datasets.features.ClassLabel(names=label_names)}
            ),
            homepage="HOMEPAGE",
            citation=_CITATION,
            task_templates=[TextClassification(text_column=text_column, label_column=label_column)],
        )

    def _split_generators(self, dl_manager):
        """
        Return SplitGenerators.
        """
        train_path = dl_manager.download_and_extract(_DOWNLOAD_URLS["train"])
        test_path = dl_manager.download_and_extract(_DOWNLOAD_URLS["test"])

        return [
            datasets.SplitGenerator(name=datasets.Split.TRAIN, gen_kwargs={"filepath": train_path}),
            datasets.SplitGenerator(name=datasets.Split.TEST, gen_kwargs={"filepath": test_path}),
        ]

    # TODO
    def _generate_examples(self, filepath):
        """
        Per each file_path read the csv file and iterate it.
        For each row yield a tuple of (id, {"text": ..., "label": ..., ...})
        Each call to this method yields an output like below:
        ```
        (123, {"text": "I liked it", "label": "positive"})
        ```
        """
        label2id = self.info.features[self.info.task_templates[0].label_column].str2int
        logger.info("⏳ Generating examples from = %s", filepath)
        with open(filepath, encoding="utf-8") as csv_file:
            csv_reader = csv.reader(csv_file, quotechar='"', skipinitialspace=True)

            # Uncomment below line to skip the first row if your csv file has a header row
            # next(csv_reader, None)

            for id_, row in enumerate(csv_reader):
                text, label = row
                label = label2id(label)
                # Optional preprocessing here
                yield id_, {"text": text, "label": label}



from ..registry import datasets_registry  # noqa
from ..builders import build_dataset  # noqa
from .dataset_processors import *
from .data_collators import *
from .data_samplers import *
from .datasets import *



from __future__ import annotations

from functools import partial
from typing import Any, Dict, List, Literal, Optional

import numpy as np
import torch
from omegaconf import DictConfig

from ..constants import PaddingType
from .logging import Logger


__all__ = [
    "convert_batch_dict_dtype",
    "resolve_inputs_length_for_padding",
    "pad_batch_items",
    "shift_tokens_right",
    "torch2numpy",
    "get_non_numeric_keys",
    "flatten_dict",
    "set_seed",
    "dataloader_worker_init_fn",
]

logger = Logger(__name__)


def convert_batch_dict_dtype(batch_dict: dict, dtype: str = "list", skip_keys: list = None) -> dict:
    """
    Convert data dtypes of the values in a batch dict.

    Args:
        batch_dict (dict): The batched dictionary. Each key in the dict has a batch of data as its value.
        dtype (str): Target data type to convert to ("list", "numpy", "torch").
        skip_keys (list): A list of key names to skip conversion.

    Returns:
        dict: The same dict with cast values.
    """
    import numpy as np
    import torch

    skip_keys = skip_keys or []

    for key, value in batch_dict.items():
        # Skip conversion for specified keys
        if key in skip_keys:
            continue

        # Convert to python lists
        if dtype == "list":
            if isinstance(value, (torch.Tensor, np.ndarray)):
                batch_dict[key] = value.tolist()
        # Convert to numpy arrays
        elif dtype == "numpy":
            if isinstance(value, torch.Tensor):
                batch_dict[key] = value.numpy()
            elif isinstance(value, list):
                batch_dict[key] = np.array(value)
        # Convert to torch tensors
        elif dtype == "torch":
            if isinstance(value, np.ndarray):
                batch_dict[key] = torch.tensor(value)
            elif isinstance(value, list):
                batch_dict[key] = torch.tensor(value)

    return batch_dict


def resolve_inputs_length_for_padding(
    inputs: List[List[Any]],
    padding: str | PaddingType = None,
    max_length: Optional[bool | int] = None,
    truncation: Optional[bool] = True,
):
    """
    Resolve final inputs length based on padding and max_length values
    """
    inputs_max_length = max([len(x) for x in inputs])

    # Resolve padding and max_length values first
    if padding is None:
        if max_length is not None:
            padding = "max_length"
        else:
            padding = "longest"

    # Now lets resolve any conflicts
    if padding == "longest":
        if max_length is not None and max_length < inputs_max_length:
            if not truncation:
                logger.warning(
                    f"Setting padding to `longest` and a max_length={max_length} which is below the "
                    f"inputs longest value {inputs_max_length} requires truncation to be True, "
                    f"got truncation={truncation}. Falling back to truncating inputs to max_length={max_length}."
                )
                inputs_length = max_length
            else:
                inputs_length = max_length
        else:
            inputs_length = inputs_max_length

    elif padding == "max_length":
        if max_length is None:
            logger.warning(
                "Setting padding='max_length' but no max_length value is provided! Falling back to padding='longest'"
            )
            inputs_length = inputs_max_length
        else:
            if max_length < inputs_max_length and not truncation:
                logger.warning(
                    f"Cannot set max_length to {max_length} "
                    f"while max input length is {inputs_max_length} and `truncation` is `False`"
                    f"Either set `truncation=True` or increase `max_length`"
                )
                inputs_length = inputs_max_length
            else:
                inputs_length = max_length
    else:
        raise ValueError(f"Invalid padding value `{padding}`, expected either `max_length` or `longest`")

    return inputs_length


def pad_batch_items(
    inputs: List[List[int | float]],
    padding: str | PaddingType = None,
    padding_side: Literal["right", "left"] = "right",
    pad_id: int = 0,
    max_length: Optional[bool | int] = None,
    truncation: Optional[bool] = True,
):
    """
    Given a nested container of unequal sized iterables e.g, batch of token ids, pad them based on padding strategy
    Args:
        inputs: A nested iterable of unequal sized iterables (e.g, list of lists)
        padding: Padding strategy, either max_length or longest
        padding_side: Where to add padding ids, `left` or `right`, defaults to `right`
        pad_id: Pad token id, defaults to `0`
        max_length: Max input length after padding, only applicable when padding == "max_length"
        truncation: Whether to truncate if an input in the batch is longer than max_length

    Returns:
        A list of equal sized lists
    """

    inputs_length = resolve_inputs_length_for_padding(
        inputs,
        padding=padding,
        max_length=max_length,
        truncation=truncation,
    )

    padded_inputs = []
    for ids in inputs:
        difference = inputs_length - len(ids)
        if difference > 0:
            paddings = [pad_id] * difference
            padded_ids = ids + paddings if padding_side == "right" else paddings + ids
            padded_inputs.append(padded_ids)
        else:
            padded_inputs.append(ids[:inputs_length])

    return padded_inputs


def shift_tokens_right(
    token_ids: list[list[int]] | "torch.Tensor" | "np.ndarray",
    pad_token_id: int,
    decoder_start_token_id: int
):
    """
    Shift input ids one token to the right.
    """
    # Check if input is a list of lists
    if isinstance(token_ids, list):
        # Initialize shifted_input_ids with the same shape as input_ids
        shifted_input_ids = [[0] * len(row) for row in token_ids]

        for i, row in enumerate(token_ids):
            # Shift each row one token to the right
            shifted_input_ids[i][1:] = row[:-1]
            # Set the first token to decoder_start_token_id
            shifted_input_ids[i][0] = decoder_start_token_id
            # Replace any -100 values with pad_token_id
            shifted_input_ids[i] = [pad_token_id if token == -100 else token for token in shifted_input_ids[i]]
        return shifted_input_ids

    # Check if input is a NumPy array
    elif isinstance(token_ids, np.ndarray):
        # Initialize shifted_input_ids with zeros and the same shape as input_ids
        shifted_input_ids = np.zeros_like(token_ids)
        shifted_input_ids[:, 1:] = token_ids[:, :-1]
        shifted_input_ids[:, 0] = decoder_start_token_id
        # Replace any -100 values with pad_token_id
        shifted_input_ids = np.where(shifted_input_ids == -100, pad_token_id, shifted_input_ids)
        return shifted_input_ids

    # Check if input is a PyTorch tensor
    elif isinstance(token_ids, torch.Tensor):
        # Initialize shifted_input_ids with zeros and the same shape as input_ids
        shifted_input_ids = token_ids.new_zeros(token_ids.shape)
        shifted_input_ids[:, 1:] = token_ids[:, :-1].clone()
        shifted_input_ids[:, 0] = decoder_start_token_id
        # Replace any -100 values with pad_token_id
        shifted_input_ids = shifted_input_ids.masked_fill(shifted_input_ids == -100, pad_token_id)
        return shifted_input_ids

    else:
        raise TypeError("Unsupported input type. Expected list, numpy array, or torch tensor.")


def torch2numpy(*args):
    """
    Cast tensors to numpy

    Args:
        *args: Any number of torch.Tensor objects

    Returns:
        The same inputs cast to numpy
    """
    return [arg.cpu().numpy() if isinstance(arg, torch.Tensor) else arg for arg in args]


def get_non_numeric_keys(d: Dict, batched=True):
    """
    Get keys that have string values in a dictionary

    Args:
        d: The dict
        batched: Are the input dict values batched or not

    Returns:
        A list of string-valued keys
    """
    keys = []
    for k, v in d.items():
        if len(v) and isinstance(v[0], list):
            if batched and not isinstance(v[0][0], (int, float, complex)) and not isinstance(v[0][0], bool):
                keys.append(k)
            elif isinstance(v[0], str):
                keys.append(k)
    return keys


def flatten_dict(dict_config: Dict | DictConfig) -> DictConfig:
    """
    Flatten a nested Dict/DictConfig object

    Args:
        dict_config: A Dict/DictConfig object

    Returns:
        The flattened version of the dict-like object
    """

    config = DictConfig({})
    for k, v in dict_config.items():
        if isinstance(v, (Dict, DictConfig)):
            config.update(flatten_dict(v))
        else:
            config[k] = v

    return config


def set_seed(seed):
    """
    Set a global seed for all backends to handle reproducibility and determinism.
    """
    import random

    import numpy as np
    import torch

    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def dataloader_worker_init_fn(seed):
    """
    A dataloader worker init function that handles reproducibility by hard-setting the seed for all workers.
    """

    def worker_init_fn(worker_id, seed):
        set_seed(seed + worker_id)

    return partial(worker_init_fn, seed=seed)



from __future__ import annotations

from dataclasses import dataclass

from ...configs import DatasetConfig
from ...constants import Backends, PaddingType, TaskType
from ...registry import register_dataset
from ...utils import is_backend_available
from ..data_collators import SpeechRecognitionDataCollator
from .dataset import Dataset


if is_backend_available(Backends.DATASETS):
    from datasets import Audio, load_dataset

_required_backends = [Backends.LIBROSA, Backends.DATASETS]


@dataclass
class SpeechRecognitionDatasetConfig(DatasetConfig):
    name = "speech_recognition"
    task = TaskType.SPEECH_RECOGNITION
    path: str = None
    sampling_rate: int = 16000
    audio_array_padding: bool | str | PaddingType = "longest"
    max_audio_array_length: int = None
    labels_padding: bool | str | PaddingType = "longest"
    labels_max_length: int = None
    audio_file_path_column: str = "path"
    audio_column: str = "audio"
    audio_array_column: str = "array"
    transcript_column: str = "sentence"


@register_dataset("speech_recognition", config_class=SpeechRecognitionDatasetConfig)
class SpeechRecognitionDataset(Dataset):
    required_backends = _required_backends

    def __init__(self, config: SpeechRecognitionDatasetConfig, split=None, preprocessor=None, **kwargs):
        super().__init__(config, split, preprocessor=preprocessor, **kwargs)
        self.feature_extractor = self.preprocessor.audio_feature_extractor
        self.tokenizer = self.preprocessor.tokenizer
        self.data_collator = SpeechRecognitionDataCollator(
            self.feature_extractor,
            self.tokenizer,
            inputs_padding=self.config.audio_array_padding,
            inputs_max_length=self.config.max_audio_array_length,
            labels_padding=self.config.labels_padding,
            labels_max_length=self.config.labels_max_length,
        )

    def _load(self, split):
        data = load_dataset(self.config.path, split=split, cache_dir=self.cache_dir, **self.config.hf_load_kwargs)
        data = data.cast_column(self.config.audio_column, Audio(sampling_rate=self.config.sampling_rate))
        return data

    def __getitem__(self, index):
        sample_dict = self.data[index]
        transcript = sample_dict[self.config.transcript_column]
        audio_array = sample_dict[self.config.audio_column][self.config.audio_array_column]

        input_features = self.feature_extractor(
            audio_array,
            sampling_rate=self.config.sampling_rate,
            return_tensors="numpy"
        )["input_features"][0]

        labels = self.tokenizer(
            transcript,
            padding=self.config.labels_padding,
            max_length=self.config.labels_max_length,
        )

        return {
            "input_features": input_features,
            "labels": labels["token_ids"],
            "attention_mask": labels["attention_mask"],
        }



from dataclasses import dataclass

from ...configs import DatasetConfig
from ...constants import Backends, TaskType
from ...registry import register_dataset
from ...utils import Logger, is_backend_available
from ..data_collators import TextGenerationDataCollator
from .dataset import Dataset


if is_backend_available(Backends.DATASETS):
    from datasets import load_dataset

logger = Logger(__name__)


@dataclass
class TextSummarizationDatasetConfig(DatasetConfig):
    """
    Configuration class for text summarization datasets.

    Args:
        path (str): Path to the dataset.
        prefix (str): Prefix for conditional generation.
        text_field (str): Field name for text in the dataset.
        summary_field (str): Field name for summary in the dataset.
        title_field (str): Field name for title in the dataset.
        max_length (int): Maximum length of text.
        labels_max_length (int): Maximum length of the target summary.
    """

    name = "text_summarization"
    task: TaskType = TaskType.TEXT_GENERATION
    path: str = None
    prefix: str = None
    text_field: str = None
    summary_field: str = None
    title_field: str = None
    max_length: int = None
    labels_max_length: int = None


@register_dataset("text_summarization", config_class=TextSummarizationDatasetConfig)
class TextSummarizationDataset(Dataset):
    """
    A text summarization dataset class.
    As of now this class is intended for datasets existing on the Hub!

    Args:
        config (TextSummarizationDatasetConfig): Dataset config object.
        split: Which split to use.
        **kwargs: Extra config parameters to assign to the original config.
    """

    def __init__(self, config: TextSummarizationDatasetConfig, split=None, preprocessor=None, **kwargs):
        super().__init__(config, split=split, preprocessor=preprocessor, **kwargs)
        self.tokenizer = self.preprocessor.tokenizer
        self.data_collator = TextGenerationDataCollator(
            tokenizer=self.tokenizer,
            max_length=self.config.max_length,
            labels_max_length=self.config.labels_max_length,
            padding="max_length" if self.config.max_length else None,
        )

    def _load(self, split):
        """
        Load the dataset.

        Args:
            split: Dataset split.

        Returns:
            The whole dataset.

        """
        data = load_dataset(self.config.path, split=split, cache_dir=self.cache_dir, **self.config.hf_load_kwargs)
        return data

    def __getitem__(self, index):
        """
        Tokenize inputs and return a dict containing ids, masks, labels, etc.

        Args:
            index: Sample index.

        Returns:
            dict: The input data.

        """
        text = self.data[index][self.config.text_field]
        if self.config.prefix is not None:
            text = self.config.prefix + text  # for conditional generation we might need a static prefix
        summary = self.data[index][self.config.summary_field]

        inputs = self.tokenizer(
            text,
            max_length=self.config.max_length,
            padding="max_length" if self.config.max_length else None,
            return_attention_mask=True,
        )
        labels = self.tokenizer(
            summary,
            max_length=self.config.max_length,
            padding="max_length" if self.config.labels_max_length else None,
            return_attention_mask=True,
        )

        inputs["labels"] = labels["token_ids"]

        return inputs



import csv
import os

import datasets
from tqdm import tqdm


_DESCRIPTION = """\
Persian portion of the common voice 13 dataset, gathered and maintained by Hezar AI.
"""

_CITATION = """\
@inproceedings{commonvoice:2020,
  author = {Ardila, R. and Branson, M. and Davis, K. and Henretty, M. and Kohler, M. and Meyer, J. and Morais, R. and Saunders, L. and Tyers, F. M. and Weber, G.},
  title = {Common Voice: A Massively-Multilingual Speech Corpus},
  booktitle = {Proceedings of the 12th Conference on Language Resources and Evaluation (LREC 2020)},
  pages = {4211--4215},
  year = 2020
}
"""

_HOMEPAGE = "https://commonvoice.mozilla.org/en/datasets"

_LICENSE = "https://creativecommons.org/publicdomain/zero/1.0/"

_BASE_URL = "https://huggingface.co/datasets/hezarai/common-voice-13-fa/resolve/main/"

_AUDIO_URL = _BASE_URL + "audio/{split}.zip"

_TRANSCRIPT_URL = _BASE_URL + "transcripts/{split}.tsv"


class CommonVoiceFaConfig(datasets.BuilderConfig):
    """BuilderConfig for CommonVoice."""

    def __init__(self, **kwargs):
        super(CommonVoiceFaConfig, self).__init__(**kwargs)


class CommonVoice(datasets.GeneratorBasedBuilder):
    DEFAULT_WRITER_BATCH_SIZE = 1000

    BUILDER_CONFIGS = [
        CommonVoiceFaConfig(
            name="commonvoice-13-fa",
            version="1.0.0",
            description=_DESCRIPTION,
        )
    ]

    def _info(self):
        features = datasets.Features(
            {
                "client_id": datasets.Value("string"),
                "path": datasets.Value("string"),
                "audio": datasets.features.Audio(sampling_rate=48_000),
                "sentence": datasets.Value("string"),
                "up_votes": datasets.Value("int64"),
                "down_votes": datasets.Value("int64"),
                "age": datasets.Value("string"),
                "gender": datasets.Value("string"),
                "accent": datasets.Value("string"),
                "locale": datasets.Value("string"),
                "segment": datasets.Value("string"),
                "variant": datasets.Value("string"),
            }
        )

        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=features,
            supervised_keys=None,
            homepage=_HOMEPAGE,
            license=_LICENSE,
            citation=_CITATION,
            version=self.config.version,
        )

    def _split_generators(self, dl_manager):
        splits = ("train", "dev", "test")
        audio_urls = {split: _AUDIO_URL.format(split=split) for split in splits}

        archive_paths = dl_manager.download(audio_urls)
        local_extracted_archive_paths = dl_manager.extract(archive_paths) if not dl_manager.is_streaming else {}

        transcript_urls = {split: _TRANSCRIPT_URL.format(split=split) for split in splits}
        transcript_paths = dl_manager.download_and_extract(transcript_urls)

        split_generators = []
        split_names = {
            "train": datasets.Split.TRAIN,
            "dev": datasets.Split.VALIDATION,
            "test": datasets.Split.TEST,
        }
        for split in splits:
            split_generators.append(
                datasets.SplitGenerator(
                    name=split_names.get(split, split),
                    gen_kwargs={
                        "local_extracted_archive_paths": local_extracted_archive_paths.get(split),
                        "archives": [dl_manager.iter_archive(archive_paths.get(split))],
                        "transcript_path": transcript_paths[split],
                    },
                ),
            )

        return split_generators

    def _generate_examples(self, local_extracted_archive_paths, archives, transcript_path):
        data_fields = list(self._info().features.keys())
        metadata = {}
        with open(transcript_path, encoding="utf-8") as f:
            reader = csv.DictReader(f, delimiter="\t", quoting=csv.QUOTE_NONE)
            for row in tqdm(reader, desc="Reading metadata..."):
                if not row["path"].endswith(".mp3"):
                    row["path"] += ".mp3"
                # accent -> accents in CV 8.0
                if "accents" in row:
                    row["accent"] = row["accents"]
                    del row["accents"]
                # if data is incomplete, fill with empty values
                for field in data_fields:
                    if field not in row:
                        row[field] = ""
                metadata[row["path"]] = row

        for i, audio_archive in enumerate(archives):
            for path, file in audio_archive:
                _, filename = os.path.split(path)
                if filename in metadata:
                    result = dict(metadata[filename])
                    # set the audio feature and the path to the extracted file
                    path = os.path.join(local_extracted_archive_paths[i],
                                        path) if local_extracted_archive_paths else path
                    result["audio"] = {"path": path, "bytes": file.read()}
                    result["path"] = path
                    yield path, result



from dataclasses import dataclass

from ....configs import ModelConfig


@dataclass
class ViTConfig(ModelConfig):
    name = "vit"
    hidden_size: int = 768
    num_hidden_layers: int = 12
    num_attention_heads: int = 12
    intermediate_size: int = 3072
    hidden_act: str = "gelu"
    hidden_dropout_prob: float = 0.0
    attention_probs_dropout_prob: float = 0.0
    initializer_range: float = 0.02
    layer_norm_eps: float = 1e-12
    image_size: int = 224
    patch_size: int = 16
    num_channels: int = 3
    qkv_bias: bool = True
    encoder_stride: int = 16



import csv
import os

import datasets


logger = datasets.logging.get_logger(__name__)

_CITATION = """Citation"""

_DESCRIPTION = """Dataset description"""

_DOWNLOAD_URLS = {
    "train": "https://huggingface.co/datasets/hezarai/persian-license-plate-v1/resolve/main/persian_license_plate_train.csv",
    "test": "https://huggingface.co/datasets/hezarai/persian-license-plate-v1/resolve/main/persian_license_plate_test.csv",
    'train_dataset': "https://huggingface.co/datasets/hezarai/persian-license-plate-v1/resolve/main/persian_license_plate_train.zip",
    'test_dataset': "https://huggingface.co/datasets/hezarai/persian-license-plate-v1/resolve/main/persian_license_plate_test.zip",
}


class ImageToTextDatasetConfig(datasets.BuilderConfig):
    def __init__(self, **kwargs):
        super(ImageToTextDatasetConfig, self).__init__(**kwargs)


class Image2TextDatasetName(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIGS = [
        ImageToTextDatasetConfig(
            name="image to text dataset name",
            version=datasets.Version("1.0.0"),
            description=_DESCRIPTION,
        ),
    ]

    def _info(self):
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "image_path": datasets.Value("string"),
                    "label": datasets.Value("string"),
                }
            ),
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        """
        Return SplitGenerators.
        """

        train_path = dl_manager.download_and_extract(_DOWNLOAD_URLS["train"])
        test_path = dl_manager.download_and_extract(_DOWNLOAD_URLS["test"])

        archive_path = dl_manager.download(_DOWNLOAD_URLS['train_dataset'])
        train_extracted_path = dl_manager.extract(archive_path) if not dl_manager.is_streaming else ""

        archive_path = dl_manager.download(_DOWNLOAD_URLS['test_dataset'])
        test_extracted_path = dl_manager.extract(archive_path) if not dl_manager.is_streaming else ""

        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN, gen_kwargs={"filepath": train_path, "dataset_dir": train_extracted_path}
            ),
            datasets.SplitGenerator(
                name=datasets.Split.TEST, gen_kwargs={"filepath": test_path, "dataset_dir": test_extracted_path}
            ),
        ]

    def _generate_examples(self, filepath, dataset_dir):
        logger.info("⏳ Generating examples from = %s", filepath)

        with open(filepath, encoding="utf-8") as csv_file:
            csv_reader = csv.reader(csv_file, quotechar='"', skipinitialspace=True)

            # Skip header
            next(csv_reader, None)

            for id_, row in enumerate(csv_reader):
                label, filename = row
                image_path = os.path.join(dataset_dir, filename)
                yield id_, {"image_path": image_path, "label": label}



"""
A DistilBERT base language model (HuggingFace Transformers) wrapped by a Hezar Model class
"""
from __future__ import annotations

from typing import List

from ....constants import Backends
from ....models import Model
from ....registry import register_model
from ....utils import is_backend_available
from .distilbert_config import DistilBERTConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import DistilBertConfig, DistilBertModel

_required_backends = [
    Backends.TRANSFORMERS,
    Backends.TOKENIZERS,
]


@register_model("distilbert", config_class=DistilBERTConfig)
class DistilBERT(Model):
    required_backends = _required_backends

    def __init__(self, config, **kwargs):
        super().__init__(config=config, **kwargs)
        self.distilbert = DistilBertModel(DistilBertConfig(**self.config))

    def forward(
        self,
        token_ids,
        attention_mask=None,
        head_mask=None,
        inputs_embeds=None,
        output_attentions=None,
        output_hidden_states=None,
        **kwargs,
    ):
        outputs = self.distilbert(
            input_ids=token_ids,
            attention_mask=attention_mask,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=True,
        )
        return outputs

    def preprocess(self, inputs: str | List[str], **kwargs):
        if isinstance(inputs, str):
            inputs = [inputs]
        if self.preprocessor.text_normalizer is not None:
            inputs = self.preprocessor.text_normalizer(inputs)
        tokenizer = self.preprocessor.tokenizer
        inputs = tokenizer(inputs, return_tensors="torch", device=self.device)
        return inputs

    def post_process(self, model_outputs, **kwargs):
        hidden_states = model_outputs.get("hidden_states", None)
        attentions = model_outputs.get("attentions", None)
        outputs = {
            "last_hidden_state": model_outputs.get("last_hidden_state"),
            "hidden_states": hidden_states,
            "attentions": attentions,
        }
        return outputs



import csv

import datasets


logger = datasets.logging.get_logger(__name__)

_CITATION = """Citation"""

_DESCRIPTION = """Description"""

_DOWNLOAD_URLS = {
    "train": "PUT PATH TO TRAIN CSV FILE HERE (REQUIRED)",
    "dev": "PUT PATH TO DEV CSV FILE HERE (OPTIONAL)",
    "test": "PUT PATH TO TEST CSV FILE HERE (OPTIONAL BUT RECOMMENDED)",
}


class DatasetNameConfig(datasets.BuilderConfig):
    def __init__(self, **kwargs):
        super(DatasetNameConfig, self).__init__(**kwargs)


class DatasetName(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIGS = [
        DatasetNameConfig(
            name="",
            version=datasets.Version(""),
            description=_DESCRIPTION,
        ),
    ]

    def _info(self):
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "tokens": datasets.Sequence(datasets.Value("string")),
                    # TODO YOU SHOULD PUT THE EXTRACTED UNIQUE TAGS IN YOUR DATASET HERE. THIS LIST IS JUST AN EXAMPLE
                    """
                    To extract unique tags from a pandas dataframe use this code and paste the output list below.

                    ```python
                    unique_tags = df["TAGS_COLUMN_NAME"].explode().unique()
                    print(unique_tags)
                    ```
                    """
                    "pos_tags": datasets.Sequence(  # USE `pos_tags`, `ner_tags`, `chunk_tags`, etc.
                        datasets.features.ClassLabel(names=["YOU SHOULD PUT THE UNIQUE TAGS LIST HERE"])  # TODO
                    ),
                }
            ),
            homepage="PUT PATH TO THE ORIGINAL DATASET HOME PAGE HERE (OPTIONAL BUT RECOMMENDED)",
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        """
        Return SplitGenerators.
        """

        train_path = dl_manager.download_and_extract(_DOWNLOAD_URLS["train"])
        test_path = dl_manager.download_and_extract(_DOWNLOAD_URLS["test"])

        return [
            datasets.SplitGenerator(name=datasets.Split.TRAIN, gen_kwargs={"filepath": train_path}),
            datasets.SplitGenerator(name=datasets.Split.TEST, gen_kwargs={"filepath": test_path}),
        ]

    # TODO
    def _generate_examples(self, filepath):
        """
        Per each file_path read the csv file and iterate it.
        For each row yield a tuple of (id, {"tokens": ..., "tags": ..., ...})
        Each call to this method yields an output like below:
        ```
        (124, {"tokens": ["hello", "world"], "pos_tags": ["NOUN", "NOUN"]})
        ```
        """
        logger.info("⏳ Generating examples from = %s", filepath)
        with open(filepath, encoding="utf-8") as csv_file:
            csv_reader = csv.reader(csv_file, quotechar='"', skipinitialspace=True)

            # Uncomment below line to skip the first row if your csv file has a header row
            # next(csv_reader, None)

            for id_, row in enumerate(csv_reader):
                tokens, pos_tags = row
                # Optional preprocessing here
                yield id_, {"tokens": tokens, "pos_tags": pos_tags}
